/*++
"This software program is available to you under a choice of one of two
licenses.  You may choose to be licensed under either the GNU General Public
License (GPL) Version 2, June 1991, available at
http://www.fsf.org/copyleft/gpl.html, or the Intel BSD + Patent License,
the text of which follows:

"Recipient" has requested a license and Intel Corporation ("Intel") is willing
to grant a license for the software entitled InfiniBand(tm) System Software
(the "Software") being provided by Intel Corporation.

The following definitions apply to this License:

"Licensed Patents" means patent claims licensable by Intel Corporation which
are necessarily infringed by the use or sale of the Software alone or when
combined with the operating system referred to below.

"Recipient" means the party to whom Intel delivers this Software.
"Licensee" means Recipient and those third parties that receive a license to
any operating system available under the GNU Public License version 2.0 or
later.

Copyright (c) 1996-2003 Intel Corporation. All rights reserved.

The license is provided to Recipient and Recipient's Licensees under the
following terms.

Redistribution and use in source and binary forms of the Software, with or
without modification, are permitted provided that the following conditions are
met:
Redistributions of source code of the Software may retain the above copyright
notice, this list of conditions and the following disclaimer.

Redistributions in binary form of the Software may reproduce the above
copyright notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

Neither the name of Intel Corporation nor the names of its contributors shall
be used to endorse or promote products derived from this Software without
specific prior written permission.

Intel hereby grants Recipient and Licensees a non-exclusive, worldwide,
royalty-free patent license under Licensed Patents to make, use, sell, offer
to sell, import and otherwise transfer the Software, if any, in source code
and object code form. This license shall include changes to the Software that
are error corrections or other minor changes to the Software that do not add
functionality or features when the Software is incorporated in any version of
a operating system that has been distributed under the GNU General Public
License 2.0 or later.  This patent license shall apply to the combination of
the Software and any operating system licensed under the GNU Public License
version 2.0 or later if, at the time Intel provides the Software to Recipient,
such addition of the Software to the then publicly available versions of such
operating system available under the GNU Public License version 2.0 or later
(whether in gold, beta or alpha form) causes such combination to be covered by
the Licensed Patents. The patent license shall not apply to any other
combinations which include the Software. No hardware per se is licensed
hereunder.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL INTEL OR ITS CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
--*/

/*
 * Abstract:
 *	This file contains the passive lock, which synchronizes passive threads.
 *	The passive lock allows multiple readers to access a resource
 *	simultaneously, exclusive from a single thread writting.  It is similar
 *	in behavior to the passive lock, but works at passive only.
 *
 * Environment:
 *	All
 *
 * $Revision: 1.1 $
 */

#ifndef _CL_PASSIVE_LOCK_H_
#define _CL_PASSIVE_LOCK_H_

#include <complib/cl_event.h>
#include <complib/cl_atomic.h>

/****h* Component Library/Passive Lock
* NAME
*	Passive Lock
*
* DESCRIPTION
*	The Passive Lock provides synchronization between multiple threads that
*	are sharing the lock with a single thread holding the lock exclusively.
*
*	Passive lock works exclusively between threads and cannot be used in
*	situations where the caller cannot be put into a waiting state.
*
*	The passive lock functions operate a cl_plock_t structure which should
*	be treated as opaque and should be manipulated only through the provided
*	functions.
*
* SEE ALSO
*	Structures:
*		cl_plock_t
*
*	Initialization:
*		cl_plock_construct, cl_plock_init, cl_plock_destroy
*
*	Manipulation
*		cl_plock_acquire, cl_plock_excl_acquire, cl_plock_release
*********/

/****s* Component Library: Passive Lock/cl_plock_t
* NAME
*	cl_plock_t
*
* DESCRIPTION
*	Passive Lock structure.
*
*	The cl_plock_t structure should be treated as opaque and should
*	be manipulated only through the provided functions.
*
* SYNOPSIS
*/
typedef struct _cl_plock {
	cl_event_t reader_event;
	cl_event_t writer_event;
	atomic32_t reader_count;

} cl_plock_t;
/*
* FIELDS
*	reader_event
*		Event used to synchronize shared access to the lock.
*
*	writer_event
*		Event used to synchronize exclusive access to the lock.
*
*	reader_count
*		Number of threads holding the lock for shared access.
*
* SEE ALSO
*	Passive Lock
*********/

/****f* Component Library: Passive Lock/cl_plock_construct
* NAME
*	cl_plock_construct
*
* DESCRIPTION
*	The cl_plock_construct function initializes the state of a
*	passive lock.
*
* SYNOPSIS
*/
static inline void
cl_plock_construct(IN cl_plock_t * const p_lock)
{
	CL_ASSERT(p_lock);

	p_lock->reader_count = 0;
	cl_event_construct(&p_lock->reader_event);
	cl_event_construct(&p_lock->writer_event);
}

/*
* PARAMETERS
*	p_lock
*		[in] Pointer to a cl_plock_t structure whose state to initialize.
*
* RETURN VALUE
*	This function does not return a value.
*
* NOTES
*	Allows calling cl_plock_destroy without first calling cl_plock_init.
*
*	Calling cl_plock_construct is a prerequisite to calling any other
*	passive lock function except cl_plock_init.
*
* SEE ALSO
*	Passive Lock, cl_plock_init, cl_plock_destroy
*********/

/****f* Component Library: Passive Lock/cl_plock_destroy
* NAME
*	cl_plock_destroy
*
* DESCRIPTION
*	The cl_plock_destroy function performs any necessary cleanup
*	of a passive lock.
*
* SYNOPSIS
*/
static inline void
cl_plock_destroy(IN cl_plock_t * const p_lock)
{
	CL_ASSERT(p_lock);
	CL_ASSERT(p_lock->reader_count == 0);

	cl_event_destroy(&p_lock->writer_event);
	cl_event_destroy(&p_lock->reader_event);
}

/*
* PARAMETERS
*	p_lock
*		[in] Pointer to a cl_plock_t structure whose state to initialize.
*
* RETURN VALUE
*	This function does not return a value.
*
* NOTES
*	cl_plock_destroy performs any necessary cleanup of the specified
*	passive lock.
*
*	This function must only be called if cl_plock_construct or
*	cl_plock_init has been called. The passive lock must not be held
*	when calling this function.
*
* SEE ALSO
*	Passive Lock, cl_plock_construct, cl_plock_init
*********/

/****f* Component Library: Passive Lock/cl_plock_init
* NAME
*	cl_plock_init
*
* DESCRIPTION
*	The cl_plock_init function initializes a passive lock.
*
* SYNOPSIS
*/
static inline cl_status_t
cl_plock_init(IN cl_plock_t * const p_lock)
{
	cl_status_t status;

	CL_ASSERT(p_lock);

	cl_plock_construct(p_lock);

	status = cl_event_init(&p_lock->writer_event, FALSE);
	if (status != CL_SUCCESS) {
		cl_plock_destroy(p_lock);
		return (status);
	}

	status = cl_event_init(&p_lock->reader_event, FALSE);
	if (status != CL_SUCCESS) {
		cl_plock_destroy(p_lock);
		return (status);
	}

	/*
	 * Set the writer event to signalled so that the first
	 * wait operation succeeds.
	 */
	status = cl_event_signal(&p_lock->writer_event);
	if (status != CL_SUCCESS) {
		cl_plock_destroy(p_lock);
		return (status);
	}

	/*
	 * Set the reader event to signalled so that the first
	 * wait operation succeeds.
	 */
	status = cl_event_signal(&p_lock->reader_event);
	if (status != CL_SUCCESS) {
		cl_plock_destroy(p_lock);
		return (status);
	}

	return (CL_SUCCESS);
}

/*
* PARAMETERS
*	p_lock
*		[in] Pointer to a cl_plock_t structure to initialize.
*
* RETURN VALUES
*	CL_SUCCESS if the passive lock was initialized successfully.
*
*	CL_ERROR otherwise.
*
* NOTES
*	Allows calling cl_plock_acquire, cl_plock_release,
*	cl_plock_excl_acquire, and cl_plock_excl_release.
*
* SEE ALSO
*	Passive Lock, cl_plock_construct, cl_plock_destroy,
*	cl_plock_excl_acquire, cl_plock_excl_release,
*	cl_plock_acquire, cl_plock_release
*********/

/****f* Component Library: Passive Lock/cl_plock_acquire
* NAME
*	cl_plock_acquire
*
* DESCRIPTION
*	The cl_plock_acquire function acquires a passive lock for
*	shared access.
*
* SYNOPSIS
*/
static inline void
cl_plock_acquire(IN cl_plock_t * const p_lock)
{
	cl_status_t status;

	CL_ASSERT(p_lock);

	status =
	    cl_event_wait_on(&p_lock->reader_event, EVENT_NO_TIMEOUT, FALSE);
	CL_ASSERT(status == CL_SUCCESS);

	/*
	 * Increment the reader count to block a thread trying for exclusive
	 * access.
	 */
	cl_atomic_inc(&p_lock->reader_count);
#ifdef DBG_PASSIVE_LOCKS
	cl_dbg_out("cl_plock_acquire: ReaderCount = %u\n",
		   p_lock->reader_count);
#endif
	/*
	 * Release the reader event to satisfy the wait of another reader
	 * or a writer.
	 */
	cl_event_signal(&p_lock->reader_event);
}

/*
* PARAMETERS
*	p_lock
*		[in] Pointer to a cl_plock_t structure to acquire.
*
* RETURN VALUE
*	This function does not return a value.
*
* SEE ALSO
*	Passive Lock, cl_plock_release, cl_plock_excl_acquire
*********/

/****f* Component Library: Passive Lock/cl_plock_excl_acquire
* NAME
*	cl_plock_excl_acquire
*
* DESCRIPTION
*	The cl_plock_excl_acquire function acquires exclusive access
*	to a passive lock.
*
* SYNOPSIS
*/
static inline void
cl_plock_excl_acquire(IN cl_plock_t * const p_lock)
{
	cl_status_t status;

	CL_ASSERT(p_lock);

	/* Acquire the reader event.  This will block new readers. */
	status =
	    cl_event_wait_on(&p_lock->reader_event, EVENT_NO_TIMEOUT, FALSE);
	CL_ASSERT(status == CL_SUCCESS);

	/* Wait for the writer event until all readers have exited. */
	while (p_lock->reader_count) {
#ifdef DBG_PASSIVE_LOCKS
		cl_dbg_out("cl_plock_excl_acquire: ReaderCount = %u\n",
			   p_lock->reader_count);
#endif
		status =
		    cl_event_wait_on(&p_lock->writer_event, EVENT_NO_TIMEOUT,
				     FALSE);
		CL_ASSERT(status == CL_SUCCESS);
	}

#ifdef DBG_PASSIVE_LOCKS
	cl_dbg_out("cl_plock_excl_acquire: Exit\n");
#endif
}

/*
* PARAMETERS
*	p_lock
*		[in] Pointer to a cl_plock_t structure to acquire exclusively.
*
* RETURN VALUE
*	This function does not return a value.
*
* SEE ALSO
*	Passive Lock, cl_plock_release, cl_plock_acquire
*********/

/****f* Component Library: Passive Lock/cl_plock_release
* NAME
*	cl_plock_release
*
* DESCRIPTION
*	The cl_plock_release function releases a passive lock from
*	shared or exclusive access.
*
* SYNOPSIS
*/
static inline void
cl_plock_release(IN cl_plock_t * const p_lock)
{
	CL_ASSERT(p_lock);

	if (p_lock->reader_count) {

		/*
		 * Decrement the count to allow a thread waiting for exclusive
		 * access to continue.
		 */
		cl_atomic_dec(&p_lock->reader_count);

#ifdef DBG_PASSIVE_LOCKS
		cl_dbg_out("cl_plock_release: ReaderCount = %u\n",
			   p_lock->reader_count);
#endif

		/* Release a writer, if any. */
		cl_event_signal(&p_lock->writer_event);
	} else {
		/* Release threads waiting to acquire the lock. */
		cl_event_signal(&p_lock->reader_event);
		cl_event_signal(&p_lock->writer_event);

#ifdef DBG_PASSIVE_LOCKS
		cl_dbg_out("cl_plock_release: Exit\n");
#endif
	}
}

/*
* PARAMETERS
*	p_lock
*		[in] Pointer to a cl_plock_t structure to release.
*
* RETURN VALUE
*	This function does not return a value.
*
* SEE ALSO
*	Passive Lock, cl_plock_acquire, cl_plock_excl_acquire
*********/

#endif				/* _CL_PASSIVE_LOCK_H_ */
